import { Connection, PublicKey, Transaction } from "@solana/web3.js";
import { BN } from "@coral-xyz/anchor";
import {
  Context,
  FlexibleStakingInfo,
  getLockupStakingAccountOwnerOffset,
  getLockupStakingAccountUnlockedOffset,
  getWithdrawPositionClaimedOffset,
  getWithdrawPositionUserOffset,
  LOCKUP_STAKING_ACCOUNT_LEN,
  LOCKUP_VAULT_LEN,
  LockupVaultData,
  PDA,
  SCALE_FACTOR,
  SOLCAT_DECIMALS,
  STAKING_PROGRAM_ID,
  WITHDRAW_POSITION_LEN,
} from "..";
import { AccountFetcher, ParsableLockupVault, ParsableLockupStakingAccount } from "../fetcher";
import {
  createAssociatedTokenAccountInstruction,
  createTransferCheckedInstruction,
  getAccount,
  getAssociatedTokenAddressSync,
} from "@solana/spl-token";

export class StakingClient {
  ctx: Context;
  public pda: PDA;

  constructor(ctx: Context, pda: PDA) {
    this.ctx = ctx;
    this.pda = pda;
  }

  public async depositFund(amount: BN) {
    // Get config data
    const config = this.pda.config();
    const configInfo = await this.ctx.connection.getAccountInfo(config.key);
    if (!configInfo) {
      throw new Error("Config account not found");
    }
    const configData = await this.ctx.program.account.config.fetch(config.key);

    const tx = new Transaction();
    const senderATA = getAssociatedTokenAddressSync(configData.mint, this.ctx.wallet.publicKey);
    const senderATAInfo = await this.ctx.connection.getAccountInfo(senderATA);
    if (!senderATAInfo) {
      throw new Error("Sender associated token account not found");
    }
    const senderATAData = await getAccount(this.ctx.connection, senderATA);
    if ((new BN(senderATAData.amount.toString())).lt(amount)) {
      throw new Error("Insufficient token balance to deposit");
    }
    const configATA = getAssociatedTokenAddressSync(configData.mint, config.key, true);
    const configATAInfo = await this.ctx.connection.getAccountInfo(configATA);
    if (!configATAInfo) {
      tx.add(
        createAssociatedTokenAccountInstruction(
          this.ctx.wallet.publicKey,
          configATA,
          config.key,
          configData.mint,
        )
      );
    }
    tx.add(
      createTransferCheckedInstruction(
        senderATA,
        configData.mint,
        configATA,
        this.ctx.wallet.publicKey,
        BigInt(amount.toString()),
        SOLCAT_DECIMALS,
      )
    );
    return tx;
  }

  // Admin functions
  public async updateAdmin(newAdmin: PublicKey) {
    const config = this.pda.config();
    const configData = await this.ctx.fetcher.getConfig(config.key, true);
    if (!configData) {
      throw new Error("Config account not found");
    }

    if (this.ctx.wallet.publicKey.toBase58() !== configData.admin.toBase58()) {
      throw new Error("Unauthorized");
    }

    const tx = new Transaction();
    const updateAdminIx = await this.ctx.methods.updateAdmin({
      inputs: {
        newAdmin,
      },
    });
    if (!updateAdminIx.ix) {
      throw new Error("Failed to create updating admin instruction");
    }
    tx.add(updateAdminIx.ix);

    return tx;
  }

  public async withdrawFund(amount: BN) {
    const config = this.pda.config();
    const configInfo = await this.ctx.connection.getAccountInfo(config.key);
    if (!configInfo) {
      throw new Error("Config account not found");
    }
    const configData = await this.ctx.program.account.config.fetch(config.key);

    if (this.ctx.wallet.publicKey.toBase58() !== configData.admin.toBase58()) {
      throw new Error("Unauthorized");
    }

    const tx = new Transaction();
    const adminATA = getAssociatedTokenAddressSync(configData.mint, this.ctx.wallet.publicKey);
    const adminATAInfo = await this.ctx.connection.getAccountInfo(adminATA);
    if (!adminATAInfo) {
      tx.add(
        createAssociatedTokenAccountInstruction(
          this.ctx.wallet.publicKey,
          adminATA,
          this.ctx.wallet.publicKey,
          configData.mint,
        )
      );
    }
    const configATA = getAssociatedTokenAddressSync(configData.mint, config.key, true);
    const configATAInfo = await this.ctx.connection.getAccountInfo(configATA);
    if (!configATAInfo) {
      throw new Error("Config associated token account not found");
    }
    const configATAData = await getAccount(this.ctx.connection, configATA);
    if ((new BN(configATAData.amount.toString())).lt(amount)) {
      throw new Error("Insufficient token balance to withdraw");
    }
    const withdrawFundIx = await this.ctx.methods.withdrawFund({
      inputs: {
        amount,
      },
    });
    if (!withdrawFundIx.ix) {
      throw new Error("Failed to create withdrawing fund instruction");
    }
    tx.add(withdrawFundIx.ix);
    return tx;
  }

  public async updateLockupVault(index: number, apy: number, lockupDuration: BN, minStakeAmount: BN) {
    const config = this.pda.config();
    const configInfo = await this.ctx.connection.getAccountInfo(config.key);
    if (!configInfo) {
      throw new Error("Config account not found");
    }
    const configData = await this.ctx.program.account.config.fetch(config.key);
    if (this.ctx.wallet.publicKey.toBase58() !== configData.admin.toBase58()) {
      throw new Error("Unauthorized");
    }

    const lockupVault = this.pda.lockupVault(new BN(index));
    const lockupVaultInfo = await this.ctx.connection.getAccountInfo(lockupVault.key);
    if (!lockupVaultInfo) {
      throw new Error("Lockup vault is not initialized");
    }

    const tx = new Transaction();
    const updateLockupVaultIx = await this.ctx.methods.updateLockupVault({
      inputs: {
        index,
        apy,
        lockupDuration,
        minStakeAmount,
      },
    });
    if (!updateLockupVaultIx.ix) {
      throw new Error("Failed to create updating lockup vault instruction");
    }
    tx.add(updateLockupVaultIx.ix);
    return tx;
  }

  public async updateFlexibleVault(rewardRate: BN, withdrawalDelayTime: BN) {
    const config = this.pda.config();
    const configInfo = await this.ctx.connection.getAccountInfo(config.key);
    if (!configInfo) {
      throw new Error("Config account not found");
    }
    const configData = await this.ctx.program.account.config.fetch(config.key);
    if (this.ctx.wallet.publicKey.toBase58() !== configData.admin.toBase58()) {
      throw new Error("Unauthorized");
    }

    const flexibleVault = this.pda.flexibleVault();
    const flexibleVaultInfo = await this.ctx.connection.getAccountInfo(flexibleVault.key);
    if (!flexibleVaultInfo) {
      throw new Error("Flexible vault is not initialized");
    }

    const tx = new Transaction();
    const updateFlexibleVaultIx = await this.ctx.methods.updateFlexibleVault({
      inputs: {
        rewardRate,
        withdrawalDelayTime,
      },
    });
    if (!updateFlexibleVaultIx.ix) {
      throw new Error("Failed to create updating flexible vault instruction");
    }
    tx.add(updateFlexibleVaultIx.ix);
    return tx;
  }

  // User functions
  public async stake(amount: BN) {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      const initFlexibleAccountIx = await this.ctx.methods.initFlexibleAccount({
        accounts: {
          user: this.ctx.wallet.publicKey,
        },
        inputs: {
          bump: flexibleAccount.bump,
        },
      });
      if (!initFlexibleAccountIx.ix) {
        throw new Error("Failed to create initializing flexible account instruction");
      }
      tx.add(initFlexibleAccountIx.ix);
    }

    const stakeIx = await this.ctx.methods.stake({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        amount,
      },
    });
    if (!stakeIx.ix) {
      throw new Error("Failed to create staking instruction");
    }
    tx.add(stakeIx.ix);

    return tx;
  }

  public async withdraw(amount: BN) {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      throw new Error("Ensure user has staked before withdrawing");
    }

    const userAccount = this.pda.user(this.ctx.wallet.publicKey);
    const userAccountInfo = await this.ctx.connection.getAccountInfo(userAccount.key);
    let withdrawIndex = new BN(0);
    if (!userAccountInfo) {
      const initUserIx = await this.ctx.methods.initUser({
        accounts: {
          user: this.ctx.wallet.publicKey,
        },
        inputs: {
          bump: userAccount.bump,
        },
      });
      if (!initUserIx.ix) {
        throw new Error("Failed to create initializing user instruction");
      }
      tx.add(initUserIx.ix);
    } else {
      const userData = await this.ctx.program.account.user.fetch(userAccount.key);
      withdrawIndex = new BN(userData.latestWithdrawIndex);
    }

    const withdrawPosition = this.pda.withdrawPosition(this.ctx.wallet.publicKey, new BN(withdrawIndex));
    const withdrawPositionInfo = await this.ctx.connection.getAccountInfo(withdrawPosition.key);
    if (withdrawPositionInfo) {
      throw new Error("Withdraw position index is already taken");
    }

    const withdrawIx = await this.ctx.methods.withdraw({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        withdrawPositionBump: withdrawPosition.bump,
        index: withdrawIndex.toNumber(),
        amount,
      },
    });
    if (!withdrawIx.ix) {
      throw new Error("Failed to create withdrawing instruction");
    }
    tx.add(withdrawIx.ix);

    return tx;
  }

  public async withdrawAfterDelay(index: number) {
    const tx = new Transaction();

    const withdrawPosition = this.pda.withdrawPosition(this.ctx.wallet.publicKey, new BN(index));
    const withdrawPositionInfo = await this.ctx.connection.getAccountInfo(withdrawPosition.key);
    if (!withdrawPositionInfo) {
      throw new Error("Ensure create withdraw position before getting delay amount");
    }

    const withdrawDelayIx = await this.ctx.methods.withdrawDelay({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        index,
      },
    });
    if (!withdrawDelayIx.ix) {
      throw new Error("Failed to create withdrawing after delay instruction");
    }
    tx.add(withdrawDelayIx.ix);

    return tx;
  }

  public async claim() {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      throw new Error("Ensure user has staked before claiming");
    }

    const claimIx = await this.ctx.methods.claim({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
    });
    if (!claimIx.ix) {
      throw new Error("Failed to create claiming instruction");
    }
    tx.add(claimIx.ix);

    return tx;
  }

  public async lockup(amount: BN, vaultIndex: number) {
    const tx = new Transaction();

    const userAccount = this.pda.user(this.ctx.wallet.publicKey);
    const userAccountInfo = await this.ctx.connection.getAccountInfo(userAccount.key);
    let lockupIndex = new BN(0);
    if (!userAccountInfo) {
      const initUserIx = await this.ctx.methods.initUser({
        accounts: {
          user: this.ctx.wallet.publicKey,
        },
        inputs: {
          bump: userAccount.bump,
        },
      });
      if (!initUserIx.ix) {
        throw new Error("Failed to create initializing user instruction");
      }
      tx.add(initUserIx.ix);
    } else {
      const userData = await this.ctx.program.account.user.fetch(userAccount.key);
      lockupIndex = new BN(userData.latestLockupIndex);
    }

    const userLockupStakingAccount = this.pda.lockupStakingAccount(
      this.ctx.wallet.publicKey,
      new BN(lockupIndex),
      new BN(vaultIndex)
    );
    const userLockupStakingAccountInfo = await this.ctx.connection.getAccountInfo(userLockupStakingAccount.key);
    if (userLockupStakingAccountInfo) {
      throw new Error("Lockup staking account is already initialized");
    }

    const lockUpIx = await this.ctx.methods.lockup({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        userLockupStakingAccountBump: userLockupStakingAccount.bump,
        index: lockupIndex.toNumber(),
        vaultIndex,
        amount,
      },
    });
    if (!lockUpIx.ix) {
      throw new Error("Failed to create locking up instruction");
    }
    tx.add(lockUpIx.ix);

    return tx;
  }

  public async unlock(index: number, vaultIndex: number) {
    const tx = new Transaction();

    const unlockIx = await this.ctx.methods.unlock({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        index,
        vaultIndex,
      },
    });
    if (!unlockIx.ix) {
      throw new Error("Failed to create unlocking instruction");
    }
    tx.add(unlockIx.ix);

    return tx;
  }

  // Getter functions
  public async getUserFlexibleStakingInfo(filter?: { withdrew: boolean }): Promise<FlexibleStakingInfo> {
    const flexibleVault = this.pda.flexibleVault();
    const flexibleVaultInfo = await this.ctx.connection.getAccountInfo(flexibleVault.key);
    if (!flexibleVaultInfo) {
      throw new Error("Flexible vault account not found");
    }
    const flexibleVaultData = await this.ctx.program.account.flexibleVault.fetch(flexibleVault.key);

    const flexibleStakingAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleStakingAccountInfo = await this.ctx.connection.getAccountInfo(flexibleStakingAccount.key);
    if (!flexibleStakingAccountInfo) {
      return {
        totalStakedAmount: "0",
        currentApr: 0,
        pendingReward: "0",
        unstakePending: "0",
        redeemPending: "0",
        withdrawPositions: [],
      }
    }
    const flexibleStakingAccountData = await this.ctx.program.account.flexibleStakingAccount.fetch(flexibleStakingAccount.key);

    const now = Math.floor(Date.now() / 1000);
    const timePassed = new BN(now).sub(flexibleVaultData.lastUpdateTime);

    let rewardPerToken = new BN(0);
    if (flexibleVaultData.totalSupply.eq(new BN(0))) {
      rewardPerToken = flexibleVaultData.rewardPerTokenStored;
    } else {
      rewardPerToken = flexibleVaultData.rewardPerTokenStored.add(
        flexibleVaultData.rewardRate
          .mul(timePassed)
          .mul(SCALE_FACTOR)
          .div(flexibleVaultData.totalSupply)
      );
    }
    const pendingReward = flexibleStakingAccountData.balance
      .mul(
        rewardPerToken.sub(flexibleStakingAccountData.userRewardPerTokenPaid)
      )
      .div(SCALE_FACTOR)
      .add(flexibleStakingAccountData.reward);

    const filters: any[] = [{ dataSize: WITHDRAW_POSITION_LEN }, { memcmp: { offset: getWithdrawPositionUserOffset(), bytes: this.ctx.wallet.publicKey.toBase58() } }];
    if (filter?.withdrew !== undefined) {
      filters.push({
        memcmp: {
          offset: getWithdrawPositionClaimedOffset(),
          bytes: filter.withdrew ? "2" : "1",
        },
      });
    }

    const withdrawPositionAccounts = await this.ctx.connection.getProgramAccounts(
      STAKING_PROGRAM_ID,
      { filters }
    );
    const withdrawPositionsInfo = await Promise.all(withdrawPositionAccounts.map(async (info) => {
      const withdrawPositionInfo = await this.ctx.program.account.withdrawPosition.fetch(info.pubkey);
      return withdrawPositionInfo;
    }));
    let unstakePending = new BN(0);
    let redeemPending = new BN(0);
    const withdrawPositions = withdrawPositionsInfo.map((info) => {
      const releaseTime = info.startDelayTime.add(flexibleVaultData.withdrawalDelayTime);
      if (releaseTime.lt(new BN(now))) {
        redeemPending = redeemPending.add(info.amount);
      } else {
        unstakePending = unstakePending.add(info.amount);
      }
      return ({
        index: info.index,
        amount: info.amount.toString(),
        startDelayTime: info.startDelayTime.toString(),
        unlockTime: info.startDelayTime.add(flexibleVaultData.withdrawalDelayTime).toString(),
        claimed: info.claimed,
      });
    });
    // sort withdraw positions by index
    withdrawPositions.sort((a: any, b: any) => a.index - b.index);

    const totalFlexibleStaked = Number(BigInt(flexibleVaultData.totalSupply.toString()));
    const apr = calculateAPR(totalFlexibleStaked, Number(flexibleVaultData.rewardRate.toString()));
    return {
      totalStakedAmount: flexibleStakingAccountData.balance.toString(),
      currentApr: apr,
      pendingReward: pendingReward.toString(),
      unstakePending: unstakePending.toString(),
      redeemPending: redeemPending.toString(),
      withdrawPositions,
    }
  }

  public async getUserLockupStakingInfo(filter?: { unlocked: boolean }) {
    const fetcher = new AccountFetcher(this.ctx.connection);
    const lockupVaultAccounts = await this.ctx.connection.getProgramAccounts(
      STAKING_PROGRAM_ID,
      {
        filters: [
          {
            dataSize: LOCKUP_VAULT_LEN
          }
        ],
      }
    );
    if (lockupVaultAccounts.length === 0) {
      throw new Error("Lockup vault account not found");
    }
    const lockupVaultsInfo = await Promise.all(lockupVaultAccounts.map(async (info) => {
      const lockupVaultInfo: any = fetcher.parseAccountData(info?.account?.data, ParsableLockupVault);
      if (!lockupVaultInfo) {
        return null;
      }
      return {
        lockupVault: info.pubkey.toBase58(),
        bump: lockupVaultInfo.bump,
        index: lockupVaultInfo.index,
        apy: lockupVaultInfo.apy,
        lockupDuration: lockupVaultInfo.lockupDuration.toString(),
        minStakeAmount: lockupVaultInfo.minStakeAmount.toString(),
        totalRewardClaimed: lockupVaultInfo.totalRewardClaimed.toString(),
      };
    }));
    lockupVaultsInfo.sort((a: any, b: any) => a.index - b.index);
    const filters: any[] = [
      { dataSize: LOCKUP_STAKING_ACCOUNT_LEN },
      { memcmp: { offset: getLockupStakingAccountOwnerOffset(), bytes: this.ctx.wallet.publicKey.toBase58() } },
    ];
    if (filter?.unlocked !== undefined) {
      filters.push({
        memcmp: {
          offset: getLockupStakingAccountUnlockedOffset(),
          bytes: filter.unlocked ? "2" : "1",
        },
      });
    }
    const lockupStakingAccounts = await this.ctx.connection.getProgramAccounts(
      STAKING_PROGRAM_ID,
      { filters }
    );
    if (lockupStakingAccounts.length === 0) {
      return {
        totalLockedAmount: "0",
        maturedAmount: "0",
        lockupStakingAccounts: [],
      }
    }
    let totalLockedAmount = new BN(0);
    let maturedAmount = new BN(0);
    const lockupStakingAccountsInfo = await Promise.all(lockupStakingAccounts.map(async (info) => {
      const lockupStakingAccountInfo: any = fetcher.parseAccountData(info?.account?.data, ParsableLockupStakingAccount);
      if (!lockupStakingAccountInfo) {
        return null;
      }
      totalLockedAmount = totalLockedAmount.add(lockupStakingAccountInfo.lockedAmount);
      // find lockup vault info in `lockupVaultsInfo` by lockupStakingAccountInfo.lockupVault
      const lockupVaultInfo = lockupVaultsInfo.find((vault: any) =>
        vault.lockupVault === lockupStakingAccountInfo.lockupVault.toBase58());
      if (!lockupVaultInfo) {
        throw new Error("User has an invalid lockup vault");
      } else {
        const maturedTime = lockupStakingAccountInfo.lockedAt.add(new BN(lockupVaultInfo.lockupDuration.toString()));
        if (maturedTime.lte(new BN(Math.floor(Date.now() / 1000)))) {
          maturedAmount = maturedAmount.add(lockupStakingAccountInfo.lockedAmount);
        }
      }
      return {
        bump: lockupStakingAccountInfo.bump,
        index: lockupStakingAccountInfo.index,
        lockupVault: lockupStakingAccountInfo.lockupVault.toBase58(),
        lockupVaultIndex: lockupVaultInfo.index,
        committedApy: lockupStakingAccountInfo.committedApy,
        lockedAmount: lockupStakingAccountInfo.lockedAmount.toString(),
        lockedAt: lockupStakingAccountInfo.lockedAt.toString(),
        unlocked: lockupStakingAccountInfo.unlocked,
      };
    }));
    // sort lockup staking accounts info by index
    lockupStakingAccountsInfo.sort((a: any, b: any) => a.index - b.index);

    return {
      totalLockedAmount: totalLockedAmount.toString(),
      maturedAmount: maturedAmount.toString(),
      lockupStakingAccounts: lockupStakingAccountsInfo,
    }
  }
}

export async function getProtocolInfo(connection: Connection) {
  const pda = new PDA(STAKING_PROGRAM_ID);
  const fetcher = new AccountFetcher(connection);
  const config = pda.config();
  const configData = await fetcher.getConfig(config.key, true);
  if (!configData) {
    throw new Error("Config account not found");
  }

  const flexibleVault = pda.flexibleVault();
  const flexibleVaultInfo = await fetcher.getFlexibleVault(flexibleVault.key, true);
  if (!flexibleVaultInfo) {
    throw new Error("Flexible vault account not found");
  }

  const lockupVaultAccounts = await connection.getProgramAccounts(
    STAKING_PROGRAM_ID,
    {
      filters: [
        {
          dataSize: LOCKUP_VAULT_LEN
        }
      ],
    }
  );
  if (lockupVaultAccounts.length === 0) {
    throw new Error("Lockup vault account not found");
  }
  const lockupVaultsInfo: LockupVaultData[] = await Promise.all(lockupVaultAccounts.map(async (info) => {
    const lockupVaultInfo = fetcher.parseAccountData(info?.account?.data, ParsableLockupVault);
    return lockupVaultInfo as LockupVaultData;
  }));
  // sort lockup vaults info by index
  lockupVaultsInfo.sort((a: any, b: any) => a.index - b.index);

  const totalFlexibleStaked = Number(BigInt(flexibleVaultInfo.totalSupply.toString()));
  const apr = calculateAPR(totalFlexibleStaked, Number(flexibleVaultInfo.rewardRate.toString()));

  return {
    admin: configData.admin.toBase58(),
    mint: configData.mint.toBase58(),
    flexibleApr: apr,
    flexibleWithdrawalDelayTime: flexibleVaultInfo.withdrawalDelayTime.toString(),
    totalFlexibleStaked: totalFlexibleStaked.toString(),
    totalFlexibleReward: flexibleVaultInfo.totalRewardClaimed.toString(),
    ...lockupVaultsInfo.reduce<{
      lockupApy: number[];
      lockupDuration: string[];
      minLockupAmount: string[];
      totalLockupStaked: string[];
      totalLockupReward: string[];
    }>((acc, vault) => {
      if (vault) {
        acc.lockupApy.push(Number(vault.apy));
        acc.lockupDuration.push(vault.lockupDuration.toString());
        acc.minLockupAmount.push(vault.minStakeAmount.toString());
        acc.totalLockupStaked.push(vault.totalLockedAmount.toString());
        acc.totalLockupReward.push(vault.totalRewardClaimed.toString());
      }
      return acc;
    }, {
      lockupApy: [],
      lockupDuration: [],
      minLockupAmount: [],
      totalLockupStaked: [],
      totalLockupReward: [],
    }),
  };
}

function calculateAPR(totalStaked: number, rewardPerSecond: number) {
  if (totalStaked === 0) return 0;
  const secondsInYear = 365 * 24 * 60 * 60;
  const annualReward = rewardPerSecond * secondsInYear;
  const apr = (annualReward / totalStaked) * 100;
  return apr;
}
