import { accountsCoder, AccountName, FlexibleVaultData, LockupVaultData, LockupStakingAccountData, ConfigData } from "../types";

/**
 * Static abstract class definition to parse entities.
 * @category Parsable
 */
export interface ParsableEntity<T> {
  /**
   * Parse account data
   *
   * @param accountData Buffer data for the entity
   * @returns Parsed entity
   */
  parse: (accountData: Buffer | undefined | null) => T | null;
}

/**
 * Class decorator to define an interface with static methods
 * Reference: https://github.com/Microsoft/TypeScript/issues/13462#issuecomment-295685298
 */
function staticImplements<T>() {
  return <U extends T>(constructor: U) => {
    constructor;
  };
}

function parseAnchorAccount(accountName: AccountName, data: Buffer) {
  const discriminator = accountsCoder.accountDiscriminator(accountName);
  if (discriminator.compare(data.subarray(0, 8))) {
    console.error("incorrect account name during parsing");
    return null;
  }

  try {
    return accountsCoder.decode(accountName, data);
  } catch (_e) {
    console.error("unknown account name during parsing");
    return null;
  }
}

// YOUR ACCOUNTS
@staticImplements<ParsableEntity<ConfigData>>()
export class ParsableConfig {
  private constructor() { }

  public static parse(data: Buffer | undefined | null): ConfigData | null {
    if (!data) {
      return null;
    }

    try {
      return parseAnchorAccount(AccountName.Config, data);
    } catch (e) {
      console.error(`error while parsing Config: ${e}`);
      return null;
    }
  }
}

@staticImplements<ParsableEntity<FlexibleVaultData>>()
export class ParsableFlexibleVault {
  private constructor() { }

  public static parse(data: Buffer | undefined | null): FlexibleVaultData | null {
    if (!data) {
      return null;
    }

    try {
      return parseAnchorAccount(AccountName.FlexibleVault, data);
    } catch (e) {
      console.error(`error while parsing Flexible Vault: ${e}`);
      return null;
    }
  }
}

@staticImplements<ParsableEntity<LockupVaultData>>()
export class ParsableLockupVault {
  private constructor() { }

  public static parse(data: Buffer | undefined | null): LockupVaultData | null {
    if (!data) {
      return null;
    }

    try {
      return parseAnchorAccount(AccountName.LockupVault, data);
    } catch (e) {
      console.error(`error while parsing Lockup Vault: ${e}`);
      return null;
    }
  }
}

@staticImplements<ParsableEntity<LockupStakingAccountData>>()
export class ParsableLockupStakingAccount {
  private constructor() { }

  public static parse(data: Buffer | undefined | null): LockupStakingAccountData | null {
    if (!data) {
      return null;
    }

    try {
      return parseAnchorAccount(AccountName.LockupStakingAccount, data);
    } catch (e) {
      console.error(`error while parsing Lockup Staking Account: ${e}`);
      return null;
    }
  }
}
